@@ -34,10 +34,10 @@ from pairtree_object import PairtreeStorageObject
 class PairtreeStorageClient(object):
     """A client that oversees the implementation of the Pairtree FS specification
     version 0.1.
-    
+
     >>> from pairtree import PairtreeStorageClient
     >>> store = PairtreeStorageClient(store_dir='data', uri_base="http://")
-    
+
     This will create the following on disc in a directory called 'data' if it doesn't already exist::
 
         $ ls -R data/
@@ -46,10 +46,10 @@ class PairtreeStorageClient(object):
 
         data/pairtree_root:
 
-    Where 
+    Where
         1. the file 'pairtree_prefix' contains just "http://"
         2. the file 'pairtree_version0_1' contains::
-        
+
            This directory conforms to Pairtree Version 0.1.
            Updated spec: http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html
 
@@ -79,9 +79,9 @@ class PairtreeStorageClient(object):
 
     def id_encode(self, id):
         """
-        The identifier string is cleaned of characters that are expected to occur rarely 
-        in object identifiers but that would cause certain known problems for file systems. 
-        In this step, every UTF-8 octet outside the range of visible ASCII (94 characters 
+        The identifier string is cleaned of characters that are expected to occur rarely
+        in object identifiers but that would cause certain known problems for file systems.
+        In this step, every UTF-8 octet outside the range of visible ASCII (94 characters
         with hexadecimal codes 21-7e) [ASCII] (Cerf, “ASCII format for network interchange,”
         October 1969.), as well as the following visible ASCII characters::
 
@@ -90,8 +90,8 @@ class PairtreeStorageClient(object):
            +   hex 2b           >   hex 3e           |   hex 7c
            ,   hex 2c
 
-        must be converted to their corresponding 3-character hexadecimal encoding, ^hh, 
-        where ^ is a circumflex and hh is two hex digits. For example, ' ' (space) is 
+        must be converted to their corresponding 3-character hexadecimal encoding, ^hh,
+        where ^ is a circumflex and hh is two hex digits. For example, ' ' (space) is
         converted to ^20 and '*' to ^2a.
 
         In the second step, the following single-character to single-character conversions
@@ -101,10 +101,10 @@ class PairtreeStorageClient(object):
                : -> +
                . -> ,
 
-        These are characters that occur quite commonly in opaque identifiers but present 
-        special problems for filesystems. This step avoids requiring them to be hex encoded 
-        (hence expanded to three characters), which keeps the typical ppath reasonably 
-        short. Here are examples of identifier strings after cleaning and after 
+        These are characters that occur quite commonly in opaque identifiers but present
+        special problems for filesystems. This step avoids requiring them to be hex encoded
+        (hence expanded to three characters), which keeps the typical ppath reasonably
+        short. Here are examples of identifier strings after cleaning and after
         ppath mapping::
 
             id:  ark:/13030/xt12t3
@@ -118,50 +118,62 @@ class PairtreeStorageClient(object):
                 ->  wh/at/-t/he/-^/2a/@^/3f/#!/^5/e!/^3/f/
 
         (From section 3 of the Pairtree specification)
-        
+
         @param id: Encode the given identifier according to the pairtree 0.1 specification
         @type id: identifier
         @returns: A string of the encoded identifier
         """
-        multichar_mapping = {'"':'^22',
-                             '<':'^3c',
-                             '?':'^3f',
-                             '*':'^2a',
-                             '=':'^3d',
-                             '^':'^5e',
-                             '+':'^2b',
-                             '>':'^3e',
-                             '|':'^7c',
-                             ',':'^2c'
+        multichar_mapping = {u'"':u'^22',
+                             u'<':u'^3c',
+                             u'?':u'^3f',
+                             u'*':u'^2a',
+                             u'=':u'^3d',
+                             u'^':u'^5e',
+                             u'+':u'^2b',
+                             u'>':u'^3e',
+                             u'|':u'^7c',
+                             u',':u'^2c'
                             }
+        second_pass_m = {u'/':u'=',
+                         u':':u'+',
+                         u'.':u','
+                        }
         new_id = []
         for char in id:
             new_id.append(multichar_mapping.get(char, char))
-        return "".join(new_id).translate(string.maketrans('/:.','=+,'))
-    
+        # 2nd pass
+        # Ditched using .translate and string.maketrans due to weird and
+        # odd errors
+        second_pass = []
+        for char in u"".join(new_id):
+            second_pass.append(second_pass_m.get(char, char))
+        return u"".join(second_pass)
+        #        m = string.maketrans(u'/:.',u'=+,')
+        #        return foo.translate(m)
+        # return "".join(new_id).translate(string.maketrans('/:.','=+,'))
+
     def id_decode(self, id):
         """
-        This decodes a given identifier from its pairtree filesystem encoding, into 
+        This decodes a given identifier from its pairtree filesystem encoding, into
         its original form:
         @param id: Identifier to decode
         @type id: identifier
         @returns: A string of the decoded identifier
         """
-        multichar_mapping = {'22':'"',
-                             '3c':'<',
-                             '3f':'?',
-                             '2a':'*',
-                             '3d':'=',
-                             '5e':'^',
-                             '2b':'+',
-                             '3e':'>',
-                             '7c':'|',
-                             '2c':','
+        multichar_mapping = {u'22':u'"',
+                             u'3c':u'<',
+                             u'3f':u'?',
+                             u'2a':u'*',
+                             u'3d':u'=',
+                             u'5e':u'^',
+                             u'2b':u'+',
+                             u'3e':u'>',
+                             u'7c':u'|',
+                             u'2c':u','
                             }
-        dec_id = id.translate(string.maketrans('=+,','/:.'))
+        dec_id = id.translate(string.maketrans(u'=+,',u'/:.'))
         index = 0
         new_id = []
-        print len(dec_id)
         while index < len(dec_id):
             if dec_id[index] == '^':
                 code = "".join(dec_id[index+1:index+3]).lower()
@@ -179,12 +191,12 @@ class PairtreeStorageClient(object):
         """
         Internal - method for discovering the pairtree identifier for a
         given directory path.
-        
+
         E.g.  pairtree_root/fo/ob/ar/+/  --> 'foobar:'
-        
+
         @param dirpath: Directory path to decode
         @type dirpath: Path to object's root
-        @returns: Decoded identifier 
+        @returns: Decoded identifier
         """
         path = self._get_path_from_dirpath(dirpath)
         return self.id_decode("".join(path))
@@ -193,7 +205,7 @@ class PairtreeStorageClient(object):
         """
         Internal - walks a directory chain and builds a list of the directory shorties
         relative to the pairtree_root
-        
+
         @param dirpath: Directory path to walk
         @type dirpath: Directory path
         """
@@ -205,21 +217,37 @@ class PairtreeStorageClient(object):
         path.reverse()
         return path
 
+
+
     def _id_to_dirpath(self, id):
         """
         Internal - method for turning an identifier into a pairtree directory tree
         of shorties.
-        
+
             -  I{"foobar://ark.1" --> "fo/ob/ar/+=/ar/k,/1"}
-        
+
         @param id: Identifer for a pairtree object
         @type id: identifier
         @returns: A directory path to the object's root directory
         """
+        return os.sep.join(self._id_to_dir_list(id))
+
+
+    def _id_to_dir_list(self, id):
+        """
+        Internal - method for turning an identifier into a list of pairtree 
+        directory tree of shorties.
+
+            -  I{"foobar://ark.1" --> ["fo","ob","ar","+=","ar","k,","1"]}
+
+        @param id: Identifer for a pairtree object
+        @type id: identifier
+        @returns: A list of directory path fragments to the object's root directory
+        """
         enc_id = self.id_encode(id)
-        dirpath = self.pairtree_root
+        dirpath = [self.pairtree_root]
         while enc_id:
-            dirpath = os.path.join(dirpath, enc_id[:self.shorty_length])
+            dirpath.append(enc_id[:self.shorty_length])
             enc_id = enc_id[self.shorty_length:]
         return dirpath
 
@@ -227,7 +255,7 @@ class PairtreeStorageClient(object):
         """
         Initialise the store if the directory doesn't exist. Create the basic structure
         needed and write the prefix to disc.
-        
+
         If the store directory exists, one of two things can happen:
             1. If that directory can be understood by this library as a pairtree store,
                it will attempt to read in the correct pairtree_prefix to use, instead of
@@ -269,7 +297,7 @@ class PairtreeStorageClient(object):
         store. This will return objects in 'split-ends' and will function correctly
         as long as non-shortie directorys are just that; non-shortie directories must
         have longer labels than the shorties - e.g::
-        
+
               ab -- cd -- ef -- foo.txt
                      |     |
                      |     ---- gh
@@ -277,21 +305,21 @@ class PairtreeStorageClient(object):
                      |           ---- foo.txt
                      |
                      ---- e  -- foo.txt
-              
+
               This method will return ['abcdef', 'abcde', 'abcdefgh'] as ids in this
               store.
-              
+
         TODO: Need to make sure this corresponds to pairtree spec.
-        
+
         Currently, it ignores the possibility of a split end being
         'shielded' by a /obj/ folder
-         
+
         (Note - this will shortly be turned into a generator, due to the expense of
         the call)
-        
+
         @returns: L{list}
         """
-         
+
         objects = set()
         paths = [os.path.join(self.pairtree_root, x) for x in os.listdir(self.pairtree_root) if os.path.isdir(os.path.join(self.pairtree_root, x))]
         d = None
@@ -313,30 +341,30 @@ class PairtreeStorageClient(object):
         """
         Internal - create an object. If the object already exists, raise a
         L{ObjectAlreadyExistsException}
-        
+
         @param id: Identifier to be created
         @type id: identifier
         @returns: L{PairtreeStorageObject}
         """
-        dirpath = self._id_to_dirpath(id)
+        dirpath = os.path.join(self._id_to_dirpath(id))
         if not os.path.exists(dirpath):
             os.makedirs(dirpath)
         else:
             raise ObjectAlreadyExistsException
         return PairtreeStorageObject(id, self)
 
-    def list_parts(self, id, path):
+    def list_parts(self, id, path=None):
         """
         List all the parts of the given identifer's parts (excluding shortie directories
         belonging to other objects)
-        
+
         If path is supplied, the parts in that subdirectory are returned.
-        
+
         If the subpath doesn't exist, a L{ObjectNotFoundException} will be raised.
-        
+
         >>> store.list_parts('foobar:1', 'data/images')
         [ 'image001.tif', 'image....    ]
-        
+
         @param id: Identifier for pairtree object
         @type id: identifier
         @param path: (Optional) List the parts contained in C{path}'s subdirectory
@@ -353,10 +381,10 @@ class PairtreeStorageClient(object):
     def put_stream(self, id, path, stream_name, bytestream, buffer_size = 1024 * 8):
         """
         Store a stream of bytes into a file within a pairtree object.
-        
+
         Can be either a string of bytes, or a filelike object which supports
         bytestream.read(buffer_size) - useful for very large files.
-        
+
         @param id: Identifier for the pairtree object to write to
         @type id: identifier
         @param path: (Optional) subdirectory path to store file in
@@ -389,26 +417,26 @@ class PairtreeStorageClient(object):
         finally:
             f.close()
 
-    def get_stream(self, id, path, stream_name, streamable=True):
+    def get_stream(self, id, path, stream_name, streamable=False):
         """
         Reads a file from a pairtree object - If streamable is set to True,
         this returns the filehandle for that file, which must be C{close()}'d
         once finished with. In python 2.6 and above, this can be done easily:
-        
+
         >>> with store.get_stream('foobar:1','data/images', 'image001.tif', True) as stream:
                 # Do something with the C{stream} handle
                 pass
-        
+
         stream is closed at the end of a C{with} block
-        
+
         @param id: Identifier for the pairtree object to read from
         @type id: identifier
         @param path: (Optional) subdirectory path to retrieve file from
         @type path: Directory path
         @param stream_name: Name of the file to read in
         @type stream_name: filename
-        @param streamable: If True, returns a filelike handle to C{read()} from - 
-        I{remember to C{close()} the file!} If False, reads in the file into a 
+        @param streamable: If True, returns a filelike handle to C{read()} from -
+        I{remember to C{close()} the file!} If False, reads in the file into a
         bytestring and return that instead.
         @type streamable: True|False
         @returns: Either L{file} or L{str}
@@ -443,25 +471,86 @@ class PairtreeStorageClient(object):
             raise PartNotFoundException(id=id, path=path, stream_name=stream_name,file_path=file_path)
         os.remove(file_path)
 
-    def exists(self, id):
+    def del_path(self, id, path, recursive=False):
         """
-        Answers the question "Does object 'xxxxxxx' exist?"
-        
+        Delete a subpath from an object, and can do so recursively (optional)
+        If the path is found to be not "empty" (ie has not parts in it) and
+        recursive is not True, then it will raise a L{PathIsNotEmptyException}
+        @param id: Identifier for the pairtree object to delete from
+        @type id: identifier
+        @param path: subdirectory path to delete
+        @type path: Directory path
+        @param recursive: Whether the delete is recursive (think rm -r)
+        @type recursive: bool
+        """
+        dirpath = os.path.join(self._id_to_dirpath(id), path)
+        if not os.path.exists(dirpath):
+            raise PartNotFoundException
+        if os.path.isfile(dirpath):
+            os.remove(dirpath)
+        else:
+            all_parts = os.listdir(dirpath)
+            deletable_parts = [x for x in all_parts if len(x)>self.shorty_length]
+            if len(all_parts) == 0:
+                os.rmdir(dirpath)
+            elif recursive:
+                for item in deletable_parts:
+                    if os.path.isdir(item):
+                        shutil.rmtree(os.path.join(dirpath, item))
+                    else:
+                        os.remove(os.path.join(dirpath, item))
+                if len(all_parts) == len(deletable_parts):
+                    os.rmdir(dirpath)
+            elif len(deletable_parts) == 0:
+                # Directory not physically empty, but empty of parts
+                pass
+            else:
+                raise PathIsNotEmptyException
+
+    def delete_object(self, id):
+        """
+        Delete's an object from the pairtree store, including any parts and subpaths
+        There is no undo...
+        @param id: Identifier of the object to delete
+        @type id: identifier
+        """
+        dirs = self._id_to_dir_list(id)
+        dirpath = os.path.join(os.sep.join(dirs))
+        if not os.path.exists(dirpath):
+            raise ObjectNotFoundException
+        for item in self.list_parts(id):
+            self.del_path(id,item, recursive=True)
+        if not os.listdir(dirpath):
+            os.rmdir(dirpath)
+        # recursively delete up, if the directory is empty
+        leaf = dirs.pop()
+        while (not os.listdir(os.sep.join(dirs)) and os.sep.join(dirs) != self.pairtree_root):
+            os.rmdir(os.sep.join(dirs))
+            dirs.pop()
+
+    def exists(self, id, path=None):
+        """
+        Answers the question "Does object or object subpath/file 'xxxxxxx' exist?"
+
         @param id: Identifier for the pairtree object to look for
         @type id: identifier
+        @param path: Subpath or subfilepath to check
+        @type path: Directory path
         @returns: L{bool}
         """
         dirpath = os.path.join(self._id_to_dirpath(id))
+        if path:
+            dirpath = os.path.join(self._id_to_dirpath(id), path)
         return os.path.exists(dirpath)
 
     def _get_new_id(self):
         """
         Inbuilt method to randomly generate an id, if one is not given to either
         L{get_object} or L{create_object}.
-        
+
         Simply returns a random 14 digit long (base 10) number, not fantastically useful
         but at least makes sure it is unique in the store.
-        
+
         @returns: Random but unique 14-digit long id number
         """
         id = "%.14d" % random.randint(0,99999999999999)
@@ -472,9 +561,9 @@ class PairtreeStorageClient(object):
     def get_object(self, id=None, create_if_doesnt_exist=True):
         """
         Returns an pairtree object with identifier C{id} if it exists.
-        
+
         If the object at C{id} doesn't exist then depending on C{create_if_doesnt_exist},
-        
+
         >>> bar = client.get_object('bar')
         # the object with id 'bar' will be retrieved and created if necessary.
 
@@ -487,10 +576,10 @@ class PairtreeStorageClient(object):
         pairtree.storage_exceptions.ObjectNotFoundException
 
         (note that fake = client.get_object('doesnotexist', False) is equivalent to the above line)
-        
+
         @param id: Identifier for the pairtree object to get (or create)
         @type id: identifier
-        @param create_if_doesnt_exist: Flag - if True, an object will be created if it 
+        @param create_if_doesnt_exist: Flag - if True, an object will be created if it
         doesn't yet exist. Will raise an L{ObjectNotFoundException} if set to False
         and the object is non-existent.
         @type create_if_doesnt_exist: True|False
@@ -509,18 +598,18 @@ class PairtreeStorageClient(object):
     def create_object(self, id):
         """
         Creates a new object with identifier C{id}
-        
+
         >>> bar = client.create_object('bar')
         >>>
-        
+
         Note that reissuing that command again will raise an L{ObjectAlreadyExistsException}:
-        
+
         >>> bar = client.create_object('bar')
         Traceback (most recent call last):
           File "<stdin>", line 1, in <module>
           File "build/bdist.linux-i686/egg/pairtree/pairtree_client.py", line 235, in create_object
         pairtree.storage_exceptions.ObjectAlreadyExistsException
-        
+
         @param id: Identifier for the pairtree object to create
         @type id: identifier
         @returns: L{PairtreeStorageObject}
